import re,sys

newline=re.compile(r"[\r\n]+")

def choose(n,k):
  if k==0:
    return [([],range(n))]
  else:
    return [(y[0]+[i],y[1]+range(i+1,n)) for i in range(k-1,n) \
                                         for y in choose(i,k-1)]

class queue:
  def __init__(self):
    self.event_map={}
    self.event_list=[[] for i in range(3)]
  def enqueue(self,event):
    if event.constraint in self.event_map:
      return
    self.event_map[event.constraint]=None
    self.event_list[event.priority].append(event.constraint)
    # Would have been better if things could change priority once in the queue.
  def dequeue(self):
    for i in range(3):
      if len(self.event_list[i])>0:
        rc=self.event_list[i][0]
        self.event_list[i]=self.event_list[i][1:]
        del self.event_map[rc]
        return rc
    return None

class msg:
  def __init__(self,priority,constraint):
    self.priority=priority # priority of message: 0=highest
    self.constraint=constraint # constraint that needs re-checking

class loctype:
  def __init__(self):
    self.constraints=[]

class sudoku:
  def __init__(self,n=3):
    self.n=n
    self.numbers=self.n*self.n
    self.rows=self.n*self.n
    self.columns=self.n*self.n
    self.area=self.rows*self.columns
    self.sectors=self.n*self.n
    self.constraints=[]
    for j in range(self.n):
      for i in range(self.n):
        self.constraints.append([self.pos(self.n*i+x,self.n*j+y) \
          for x in range(self.n) for y in range(self.n)])
    # The "localize" method requires sectors to be defined first.
    for i in range(self.columns):
      self.constraints.append([self.pos(i,x) for x in range(self.rows)])
    for i in range(self.rows):
      self.constraints.append([self.pos(x,i) for x in range(self.columns)])
    self.constraints_mask=[0 for i in range(len(self.constraints))]
    for i in range(len(self.constraints)):
      for j in self.constraints[i]:
        self.constraints_mask[i]+=2**j
    if n==3:
      self.constraint_names=['sector alpha','sector beta','sector gamma',\
                        'sector delta','sector epsilon','sector zeta', \
                        'sector eta','sector theta','sector iota', \
                        'column A','column B', \
                        'column C','column D','column E','column F', \
                        'column G','column H','column I', \
                        'row 9','row 8','row 7','row 6','row 5','row 4', \
                        'row 3','row 2','row 1']
      self.dictionary={"1":0,"2":1,"3":2,"4":3,"5":4,"6":5,"7":6,"8":7,"9":8}
      self.inv_dict=["1","2","3","4","5","6","7","8","9"]
    else:
      self.constraint_names=['sector '+str(i+1) for i in range(self.sectors)]+ \
                       ['column '+str(i+1) for i in range(self.columns)]+ \
                       ['row '+str(self.rows-i) for i in range(self.rows)]
    self.prepare_partials()
    self.dictionary={}
    self.inv_dict=[]
  def prepare_partials(self):
    # Prepreparation of this list enables us to pick a convenient checking
    # order as well.
    self.partials=[]
    for i in range(1,(self.numbers+1)/2):
      self.partials.extend(choose(self.numbers,i))
      self.partials.extend(choose(self.numbers,self.numbers-i))
    if self.numbers%2==0:
      self.partials.extend(choose(self.numbers,self.numbers/2))
  def pos(self,x,y):
    return y*self.columns+x
  def find_constraints(self,position):
    return [i for i in range(len(self.constraints)) \
            if position in self.constraints[i]]
  def clear(self):
    self.board=[[1]*self.numbers for i in range(self.area)]
    self.knowledge=[[0]*self.sectors for j in range(self.numbers)]
    self.pending=queue()
    if self.n!=3:
      self.dictionary={}
      self.inv_dict=[" "]*self.numbers
    self.board_explain=[["Unknown"]*self.numbers for i in range(self.area)]
    self.remaining_atoms=self.area
  def solve(self,board,verbose=False):
    self.clear()
    self.read(board)
    self.main(verbose)
    if not verbose:
      self.output()
  def read(self,board):
    distinct=0
    board=newline.sub("\n",board)
    for i in range(len(board)):
      x=i%(self.columns+1) # because of new-lines
      y=i/(self.columns+1)
      c=board[i]
      if x<self.columns and c!=' ':
        position=self.pos(x,y)
        if c not in self.dictionary:
          if distinct==self.numbers:
            print "Here are some of the elements appearing on the board:"
            print self.dictionary.keys()
            raise "Board contains too many distinct elements."
          val=distinct
          self.dictionary[c]=val
          distinct+=1
        else:
          val=self.dictionary[c]
        for i in range(self.numbers):
          if i!=val:
            self.board[position][i]=0
          self.board_explain[position][i]="while reading the board"
        for c in self.find_constraints(position):
          self.pending.enqueue(msg(0,c))
    self.inv_dict=range(self.numbers)
    for i in self.dictionary:
      self.inv_dict[self.dictionary[i]]=i
  def localize(self,poslist):
    rc=loctype()
    mask=0
    for i in poslist:
      mask+=2**i
    for i in range(len(self.constraints_mask)):
      if mask | self.constraints_mask[i]==self.constraints_mask[i]:
        rc.constraints.append(i)
    rc.exactness=len(rc.constraints)
    rc.sector=rc.constraints[0] # this is why sectors are defined first
    names=[self.constraint_names[x] for x in rc.constraints]
    if len(names)>1:
      names[-1:-1]=["and"]
    rc.constraints_names=" ".join(names)
    total={}
    for i in rc.constraints:
      for j in self.constraints[i]:
        total[j]=None
    for i in poslist:
      del total[i]
    rc.broadcast=total.keys()
    return rc
  def locnames(self,locs):
    rc=[]
    for i in locs:
      x=i%self.columns
      y=i/self.columns
      if self.n==3:
        x="ABCDEFGHI"[x]
        y="987654321"[y]
        rc.append(x+y)
      else:
        rc.append(str((x+1,self.columns-y)))
    rc.sort()
    if len(rc)>1:
      rc[-1:-1]=["and"]
    return " ".join(rc)
  def main(self,verbose=False):
    while self.remaining_atoms>0:
      constraint=self.pending.dequeue()
      if constraint==None:
        return
      transpose=[{} for i in range(self.numbers)]
      for i in range(self.numbers):
        for j in self.constraints[constraint]:
          if self.board[j][i]==1:
            transpose[i][j]=None
      for i in range(self.numbers):
        loc=self.localize(transpose[i])
        if loc.exactness>1 and self.knowledge[i][loc.sector]<loc.exactness:
          self.knowledge[i][loc.sector]=loc.exactness
          if loc.exactness==3:
            self.remaining_atoms-=1
          for position in loc.broadcast:
            if self.board[position][i]==1:
              self.board[position][i]=0
              explanation="while checking "+ \
                self.inv_dict[i]+ \
                " on "+self.constraint_names[constraint]+ \
                " and finding it only on "+ \
                self.locnames(transpose[i].keys())
              self.board_explain[position][i]=explanation
              if sum(self.board[position])==1:
                self.board_explain[position] \
                  [self.board[position].index(1)]=explanation
                if verbose:
                  self.output()
              for c in self.find_constraints(position):
                self.pending.enqueue(msg(1,c))
      for (s1,s2) in self.partials:
        total={}
        for i in s1:
          for position in transpose[i]:
            total[position]=None
        if len(total)==len(s1):
          for i in s2:
            for position in total:
              if position in transpose[i]:
                self.board[position][i]=0
                del transpose[i][position]
                if len(s1)*2<=self.numbers:
                  values=[self.inv_dict[x] for x in s1]
                  values.sort()
                  places=total.keys()
                  places.sort()
                  if len(values)==1:
                    connector="it"
                  else:
                    connector="them"
                  explanation="while checking "+ \
                    str(values)+ \
                    " on "+self.constraint_names[constraint]+ \
                    " and finding "+connector+" only on "+ \
                    self.locnames(places)
                else:
                  values=[self.inv_dict[x] for x in s2]
                  values.sort()
                  remainder=self.constraints[constraint][:]
                  for v in total:
                    remainder.remove(v)
                  if len(remainder)>1:
                    explanation="while checking "+ \
                      self.locnames(remainder)+ \
                      " on "+self.constraint_names[constraint]+ \
                      " and finding that they only contain values from "+ \
                      str(values)
                  else:
                    explanation="while checking "+ \
                      self.constraint_names[constraint]+ \
                      " and discovering that "+self.locnames(remainder)+ \
                      " is "+str(values[0])
                self.board_explain[position][i]=explanation
                if sum(self.board[position])==1:
                  self.board_explain[position] \
                    [self.board[position].index(1)]=explanation
                  if verbose:
                    self.output()
                for c in self.find_constraints(position):
                  self.pending.enqueue(msg(2,c))
  def output(self):
    rc=""
    for y in range(self.rows):
      for x in range(self.columns):
        data=self.board[self.pos(x,y)]
        if sum(data)==1:
          rc+=self.inv_dict[data.index(1)]
        else:
          rc+=' '
      rc+='\n'
    print rc
  def explain(self,position_string,value_string):
    value=self.dictionary[value_string]
    if type(position_string)==type(""):
      x=ord(position_string[0])-ord('a')
      if (x<0) or (x>=self.numbers) or position_string[0]>'z':
        x=ord(position_string[0])-ord('A')
        if (x<0) or (x>=self.numbers) or position_string[0]>'Z':
          raise "Unknown position format. Use \"D6\" or (4,6) type format."
      y=self.numbers-(ord(position_string[1])-ord('0'))
      if (y<0) or (y>=self.numbers):
        raise "Unknown position format. Use \"D6\" or (4,6) type format."
    else:
      x=position_string[0]-1
      y=self.numbers-position_string[1]
    position=self.pos(x,y)
    if self.board[position][value]==0:
      print value_string,"is not on",position_string,". This was discovered",
      print self.board_explain[position][value]+"."
    else:
      if sum(self.board[position])==1:
        print position_string,"is",value_string+". This was discovered",
        print self.board_explain[position][value]+"."
      else:
        print "It is still undetermined whether",value_string,"is on",
        print position_string,"or not. Possibilities are",
        values=[self.inv_dict[i] for i in range(self.numbers) \
                              if self.board[position][i]==1]
        values.sort()
        values[-1:-1]=["and"]
        print " ".join(values)+"."

if __name__=="__main__":
  p=sudoku()
  p.solve(sys.stdin.read())

